import WebSocket from 'ws';
import { EventEmitter } from './emitter';
import { Subscription } from './subscription';
import { TopicEmitter } from './topic';
import { SignedEvent } from '../event/SignedEvent';
import { Transformer, Middleware } from './transformer';
import { ClientConfig, Event, EventDraft, EventTemplate, Filter, ClientOptions, Tag, AckEnvelope, Sorter, TopicOptions } from '../schema/types';
type ClientTransformer = Transformer<SignedEvent, NostrClient>;
type ClientMiddleware = Middleware<SignedEvent, NostrClient>;
export declare class NostrClient extends EventEmitter<{
    'ready': [NostrClient];
    'notice': [string];
    'info': [string];
    'debug': [string];
    'error': [unknown];
    [k: string]: any;
}> {
    readonly id: string;
    readonly middleware: ClientTransformer;
    private keypair;
    address?: string;
    socket?: WebSocket;
    filter: Filter;
    options: ClientOptions;
    tags: Tag[][];
    ready: boolean;
    static defaults: {
        kind: number;
        tags: never[];
        selfsub: boolean;
        timeout: number;
        filter: {
            since: number;
        };
    };
    constructor(options?: ClientConfig);
    get initialized(): boolean;
    get connected(): boolean;
    get prvkey(): string;
    set prvkey(key: string | Uint8Array);
    get pubkey(): string;
    private _openHandler;
    private _messageHandler;
    importSeed(string: string): Promise<void>;
    use(fn: ClientMiddleware): number;
    connect(address?: string): Promise<NostrClient>;
    subscribe(filter?: Filter): Subscription;
    topic(topic: string, options?: TopicOptions): TopicEmitter;
    query(filter: Filter, sorter?: Sorter<SignedEvent>): Promise<SignedEvent[]>;
    cancel(subId: string): void;
    publish(draft: EventDraft): Promise<AckEnvelope | undefined>;
    sign(event: EventTemplate): Promise<Event>;
    send(message: string): Promise<void>;
    close(): void;
}
export {};
//# sourceMappingURL=client.d.ts.map